

class WappstoError(Exception):
    """
    Wappsto IoT Main Exception Class.

    All Wappsto IoT exceptions are derived from this class.
    """
    pass


class ConfigNotFoundError(FileNotFoundError, WappstoError):
    """Exception used when config file(s) was not found or created."""
    pass


class ConfigFileError(ConfigNotFoundError, WappstoError):
    """Exception used when there was found an error in the Config file(s)."""
    pass



from enum import Enum

from typing import Optional
from typing import Dict

from pydantic import BaseModel


# #############################################################################
#                             Value Settings Schema
# #############################################################################

class IoTEvent(str, Enum):
    CREATE = "create"  # POST
    CHANGE = "change"  # PUT
    REQUEST = "request"  # GET
    DELETE = "delete"  # DELETE


class ValueBaseType(str, Enum):
    """Internal use only!."""
    STRING = "string"
    NUMBER = "number"
    BLOB = "blob"
    XML = "xml"


class ValueSettinsSchema(BaseModel):
    value_type: ValueBaseType
    type: str
    mapping: Optional[Dict]  # Number only
    ordered_mapping: Optional[bool]  # Number only
    meaningful_zero: Optional[bool]  # Number only
    si_conversion: Optional[str]  # Number only
    min: Optional[float]  # Number only
    max: Optional[float]  # Blob, number, str only.
    step: Optional[float]  # Number only
    encoding: Optional[str]  # Blob, str only.
    xsd: Optional[str]  # XML only
    namespace: Optional[str]  # XML only
    unit: Optional[str]  # Number only


class ValueTemplate(str, Enum):
    """
    Predefined ValueTypes.

    Each of the predefined ValueTypes, have default
    value parameters set, which include BaseType, name,
    permission, range, step and the unit.
    """

    __version__ = "0.0.1"

    ANGLE = "ANGLE"
    BLOB = "BLOB"
    BOOLEAN_ONOFF = "BOOLEAN_ONOFF"
    CITY = "CITY"
    CO2 = "CO2"
    COLOR_HEX = "COLOR_HEX"
    COLOR_INT = "COLOR_INT"
    COLOR_TEMPERATURE = "COLOR_TEMPERATURE"
    COUNTRY = "COUNTRY"
    COUNTRY_CODE = "COUNTRY_CODE"
    ENERGY_KWH = "ENERGY_KWH"
    ENERGY_WH = "ENERGY_WH"
    HUMIDITY = "HUMIDITY"
    IMAGE_JPG = "IMAGE_JPG"
    LATITUDE = "LATITUDE"
    LONGITUDE = "LONGITUDE"
    LUMINOUSITY_LX = "LUMINOUSITY_LX"
    NUMBER = "NUMBER"
    PERCENTAGE = "PERCENTAGE"
    POSTCODE = "POSTCODE"
    POWER_KW = "POWER_KW"
    POWER_WATT = "POWER_WATT"
    PRECIPITATION_MM = "PRECIPITATION_MM"
    PRESSURE_HPA = "PRESSURE_HPA"
    SPEED_MS = "SPEED_MS"
    STREET = "STREET"
    STRING = "STRING"
    TEMPERATURE_CELSIUS = "TEMPERATURE_CELSIUS"
    TEMPERATURE_FAHRENHEIT = "TEMPERATURE_FAHRENHEIT"
    TEMPERATURE_KELVIN = "TEMPERATURE_KELVIN"
    TIMESTAMP = "TIMESTAMP"
    VOLTAGE_V = "VOLTAGE_V"
    XML = "XML"


valueSettings: Dict[ValueTemplate, ValueSettinsSchema] = {

    ValueTemplate.BOOLEAN_ONOFF: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="boolean",
        mapping={'0': 'off', '1': 'on'},
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="1",
        step="1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.VOLTAGE_V: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="voltage",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="250",
        step="1",
        unit="V",
        si_conversion=None,
    ),
    ValueTemplate.POWER_WATT: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="power",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="2500",
        step="1",
        unit="W",
        si_conversion=None,
    ),
    ValueTemplate.POWER_KW: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="power",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1000000",
        step="1",
        unit="kW",
        si_conversion="[W] = 1000 * [kW]",
    ),
    ValueTemplate.ENERGY_WH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100000",
        step="1",
        unit="Wh",
        si_conversion=None,
    ),
    ValueTemplate.ENERGY_KWH: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="energy",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="1000000",
        step="1",
        unit="kWh",
        si_conversion="[J] = 3600000 * [kWh]  ",
    ),
    ValueTemplate.TEMPERATURE_CELSIUS: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-30",
        max="50",
        step="1",
        unit="°C",
        si_conversion="[K] = [°C] + 273.15",
    ),
    ValueTemplate.TEMPERATURE_FAHRENHEIT: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=False,
        min="-20",
        max="120",
        step="1",
        unit="°F",
        si_conversion="[K] = ([°F] + 459.67) × 5/9 ",
    ),
    ValueTemplate.TEMPERATURE_KELVIN: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="240",
        max="320",
        step="1",
        unit="K",
        si_conversion=None,
    ),
    ValueTemplate.ANGLE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="angle",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="360",
        step="0",
        unit="°",
        si_conversion="[rad] = (180/pi) * [°]",
    ),
    ValueTemplate.PERCENTAGE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="percentage",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100",
        step="1",
        unit="%",
        si_conversion="[1] = 100 * [%]",
    ),
    ValueTemplate.SPEED_MS: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="speed",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100",
        step="1",
        unit="m/s",
        si_conversion=None,
    ),
    ValueTemplate.PRECIPITATION_MM: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="precipitation",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="100",
        step="1",
        unit="mm",
        si_conversion=None,
    ),
    ValueTemplate.HUMIDITY: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="relative_humidity",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="100",
        step="1",
        unit="%",
        si_conversion="[1] = 100 * [%]",
    ),
    ValueTemplate.CO2: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="concentration",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=True,
        min="0",
        max="3000",
        step="1",
        unit="ppm",
        si_conversion="1000000 * [ppm]",
    ),
    ValueTemplate.PRESSURE_HPA: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="pressure",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="300",
        max="1100",
        step="1",
        unit="hPa",
        si_conversion="[Pa] = [hPa]/100",
    ),
    ValueTemplate.TIMESTAMP: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="timestamp",
        max="27",
        encoding="ISO 8601",
    ),
    ValueTemplate.LUMINOUSITY_LX: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="luminousity",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="0",
        max="25000",
        step="1",
        unit="lx",
        si_conversion=None,
    ),
    ValueTemplate.COLOR_HEX: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="color",
        max="6",
        encoding="hex",
    ),
    ValueTemplate.COLOR_INT: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="color",
        max="8",
        encoding="integer",
    ),
    ValueTemplate.COLOR_TEMPERATURE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="color_temperature",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="1000",
        max="12000",
        step="1",
        unit="K",
        si_conversion=None,
    ),
    ValueTemplate.IMAGE_JPG: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="image",
        max="255",
        encoding="base64",
    ),
    ValueTemplate.LATITUDE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="latitude",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-90",
        max="90",
        step="0.000001",
        unit="°N",
        si_conversion=None,
    ),
    ValueTemplate.LONGITUDE: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="longitude",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-180",
        max="180",
        step="0.000001",
        unit="°E",
        si_conversion=None,
    ),
    ValueTemplate.STREET: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="street",
        max="100",
        encoding="",
    ),
    ValueTemplate.CITY: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="city",
        max="100",
        encoding="",
    ),
    ValueTemplate.POSTCODE: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="postcode",
        max="10",
        encoding="",
    ),
    ValueTemplate.COUNTRY: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="country",
        max="20",
        encoding="",
    ),
    ValueTemplate.COUNTRY_CODE: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="country_code",
        max="2",
        encoding="ISO 3166-1 Alpha-2",
    ),
    ValueTemplate.NUMBER: ValueSettinsSchema(
        value_type=ValueBaseType.NUMBER,
        type="number",
        mapping=None,
        ordered_mapping=None,
        meaningful_zero=None,
        min="-128",
        max="128",
        step="0.1",
        unit=None,
        si_conversion=None,
    ),
    ValueTemplate.STRING: ValueSettinsSchema(
        value_type=ValueBaseType.STRING,
        type="string",
        max="64",
        encoding="utf-8",
    ),
    ValueTemplate.BLOB: ValueSettinsSchema(
        value_type=ValueBaseType.BLOB,
        type="blob",
        max="280",
        encoding="base64",
    ),
    ValueTemplate.XML: ValueSettinsSchema(
        value_type=ValueBaseType.XML,
        type="xml",
        xsd="",
        namespace="",
    ),
}



import logging
import json
from uuid import UUID

import httpx

from pydantic import parse_obj_as
from pydantic import ValidationError

from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

from ..template import ServiceClass

from ..schema.base_schema import BlobValue
from ..schema.base_schema import Device
from ..schema.base_schema import Network
from ..schema.base_schema import NumberValue
from ..schema.base_schema import State
from ..schema.base_schema import StringValue
from ..schema.base_schema import XmlValue

from ..schema.base_schema import WappstoMetaType
from ..schema.base_schema import WappstoMethods
from ..schema.rest_schema import DeleteList

from ..utils.Exceptions import WappstoError
from ..schema.jsonrpc_schema import StreamEvents

from ..wappsto_ws import WappstoWebSocket


class RestAPI(ServiceClass):

    def __init__(
        self,
        token,
        username,
        password,
        url,
        version="2.0",
    ):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())
        self.base_url = url
        self.update_version(version)

        if token:
            self.login_header = {"X-session": token}
        else:
            if not username or not password:
                msg = "missing required positional argument(s): "
                msg += "'username' " if not username else ""
                msg += "'password' " if not password else ""
                raise TypeError(msg)
            self.session_json = {
                "username": username,
                "password": password,
                "remember_me": True
            }
            self.start_session()
        self.header = {"Accept": "application/json", **self.login_header}

        self.ws = WappstoWebSocket(
            session_id=self.header["X-session"],
            service=url
        )

    def update_version(self, version):
        self.version = version
        _vr = f"/{version}" if version else ""
        self.url = f"https://{self.base_url}/services{_vr}"

    def start_session(self):
        url = f"{self.url}/session"

        rdata = httpx.post(
            url=url,
            headers={"Content-type": "application/json"},
            data=json.dumps(self.session_json)
        )

        if not rdata.ok:
            self._log_request_error(rdata)
            return
        rjson = json.loads(rdata.text)
        self.login_header = {"x-session": rjson["meta"]["id"]}
        return rjson

    def stop_session(self):
        session_uuid = self.login_header.get('X-session')
        url = f"{self.url}/session/{session_uuid}"

        rdata = httpx.delete(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        bm_data = parse_obj_as(DeleteList, json.loads(rdata.text))
        return session_uuid in bm_data.deleted

    # #########################################################################
    #                             Helper Functions
    # #########################################################################

    def _constain_check(self, sub: Dict[Any, Any], main: Dict[Any, Any]) -> bool:
        """
        Check the the Sub-dictionary is contained inside the main-dictionary.

        Return True, if all of sub is in the main.
        Return False, if anyting in sub is missing in main.
        """
        return not bool(dict(set(sub.items()) - set(main.items())))

    _stream_event_parser: Dict[StreamEvents, WappstoMethods] = {
        StreamEvents.CREATE: WappstoMethods.POST,
        StreamEvents.UPDATE: WappstoMethods.PUT,
        StreamEvents.DELETE: WappstoMethods.DELETE,
        StreamEvents.DIRECT: ""
    }

    # #########################################################################
    #                               Network API
    # #########################################################################

    def subscribe_network_event(
        self,
        uuid: UUID,
        callback: Callable[[Network, WappstoMethods], None]
    ):
        self.ws.subscribe(
            wappsto_type=WappstoMetaType.network,
            unit_uuid=uuid,
            callback=lambda _id, data: callback(
                data.data,
                self._stream_event_parser.get(data.event)
            )
        )

    def post_network(self, data) -> bool:
        # url=f"/services/2.0/network",
        url = f"{self.url}/network"

        send = data.json()

        rdata = httpx.post(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def put_network(self, uuid: UUID, data) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)
        url = f"{self.url}/network/{uuid}"

        send = data.json()

        rdata = httpx.put(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def get_network(self, uuid: UUID) -> Network:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/network/{uuid}"

        rdata = httpx.get(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        try:
            return parse_obj_as(Network, json.loads(rdata.text))
        except ValidationError:
            raise WappstoError("Network not founded.")

    def delete_network(self, uuid: UUID) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/network/{uuid}"

        rdata = httpx.delete(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        bm_data = parse_obj_as(DeleteList, json.loads(rdata.text))
        return uuid in bm_data.deleted

    # #########################################################################
    #                                Device API
    # #########################################################################

    def subscribe_device_event(
        self,
        uuid: UUID,
        callback: Callable[[Device, WappstoMethods], None]
    ):
        self.ws.subscribe(
            wappsto_type=WappstoMetaType.device,
            unit_uuid=uuid,
            callback=lambda _id, data: callback(
                data.data,
                self._stream_event_parser.get(data.event)
            )
        )

    def post_device(self, network_uuid: UUID, data: Device) -> bool:
        # url=f"/services/2.0/network/{uuid}/device",
        url = f"{self.url}/network/{network_uuid}/device"

        send = data.json()

        rdata = httpx.post(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def put_device(self, uuid: UUID, data: Device) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)
        url = f"{self.url}/device/{uuid}"

        send = data.json()

        rdata = httpx.put(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def get_device(self, uuid: UUID) -> Device:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/device/{uuid}"

        rdata = httpx.get(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        try:
            return parse_obj_as(Device, json.loads(rdata.text))
        except ValidationError:
            raise WappstoError("Device not founded.")

    def delete_device(self, uuid: UUID) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/device/{uuid}"

        rdata = httpx.delete(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        bm_data = parse_obj_as(DeleteList, json.loads(rdata.text))
        return uuid in bm_data.deleted

    # #########################################################################
    #                                 Value API
    # #########################################################################

    ValueUnion = Union[StringValue, NumberValue, BlobValue, XmlValue]

    def subscribe_value_event(
        self,
        uuid: UUID,
        callback: Callable[[ValueUnion, WappstoMethods], None]
    ):
        self.ws.subscribe(
            wappsto_type=WappstoMetaType.value,
            unit_uuid=uuid,
            callback=lambda _id, data: callback(
                data.data,
                self._stream_event_parser.get(data.event)
            )
        )

    def post_value(self, device_uuid: UUID, data: ValueUnion) -> bool:
        # url=f"/services/2.0/device/{uuid}/value",
        url = f"{self.url}/device/{device_uuid}/value"

        send = data.json()

        rdata = httpx.post(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def put_value(self, uuid: UUID, data: ValueUnion) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)
        url = f"{self.url}/value/{uuid}"

        send = data.json()

        rdata = httpx.put(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def get_value(self, uuid: UUID) -> ValueUnion:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/value/{uuid}"

        rdata = httpx.get(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        try:
            return parse_obj_as(self.ValueUnion, json.loads(rdata.text))
        except ValidationError:
            raise WappstoError("Value not founded.")

    def delete_value(self, uuid: UUID) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/value/{uuid}"

        rdata = httpx.delete(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        bm_data = parse_obj_as(DeleteList, json.loads(rdata.text))
        return uuid in bm_data.deleted

    # #########################################################################
    #                                State API
    # #########################################################################

    def subscribe_state_event(
        self,
        uuid: UUID,
        callback: Callable[[State, WappstoMethods], None]
    ):
        self.ws.subscribe(
            wappsto_type=WappstoMetaType.sate,
            unit_uuid=uuid,
            callback=lambda _id, data: callback(
                data.data,
                self._stream_event_parser.get(data.event)
            )
        )

    def post_state(self, value_uuid: UUID, data: State) -> bool:
        # url=f"/services/2.0/{uuid}/state",
        url = f"{self.url}/value/{value_uuid}/state"

        send = data.json()

        rdata = httpx.post(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def put_state(self, uuid: UUID, data: State) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)
        url = f"{self.url}/state/{uuid}"

        send = data.json()

        rdata = httpx.put(
            url=url,
            headers=self.header,
            data=send
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        created = json.loads(rdata.text)

        return self._constain_check(sub=send, main=created)

    def get_state(self, uuid: UUID) -> Union[State, None]:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/state/{uuid}"

        rdata = httpx.get(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        try:
            return parse_obj_as(State, json.loads(rdata.text))
        except ValidationError:
            raise WappstoError("state not founded.")

    def delete_state(self, uuid: UUID) -> bool:
        if not isinstance(uuid, UUID):
            uuid = UUID(uuid)

        url = f"{self.url}/state/{uuid}"

        rdata = httpx.delete(
            url=url,
            headers=self.header
        )

        if not rdata.is_error:
            raise WappstoError("Error in Connection.")
        bm_data = parse_obj_as(DeleteList, json.loads(rdata.text))
        return uuid in bm_data.deleted



import uuid

from enum import Enum

from typing import Optional

from pydantic import BaseModel
from pydantic import Extra

from wappstoiot.schema.base_schema import BaseMeta
from wappstoiot.schema.base_schema import WappstoVersion


class WappstoObjectType(str, Enum):
    # WappstoMetaType
    NETWORK = "network"
    DEVICE = "device"
    VALUE = "value"
    STATE = "state"
    CREATOR = "creator"
    # session = "session"
    # prototype = "prototype"


class CreatorStatusType(str, Enum):
    PENDING = "pending"  # Not in use?
    ASSIGNED = "assigned"  # In use
    CREATED = "created"  # ??


class MetaAPIData(BaseMeta):
    type: WappstoObjectType


class CreatorStatusSchema(BaseModel):
    id: Optional[uuid.UUID]  # NOTE: Need to be optional.
    status: CreatorStatusType


class FactorySchema(BaseModel):
    reset_manufacturer: bool
    reset_owner: bool


class CreatorSchema(BaseModel):
    ca: Optional[str]  # NOTE: Only on reply. TODO: Change to certificates verification.
    certificate: Optional[str]  # NOTE: Only on reply. TODO: Change to certificates verification.
    private_key: Optional[str]  # NOTE: Only on reply. TODO: Change to certificates verification.
    network: Optional[CreatorStatusSchema]
    bound: Optional[bool]

    data: Optional[str]  # UNSURE: what is in this?  str of 'NetworkSchema' Only vr 2.0?
    product: Optional[str]

    factory_reset: Optional[FactorySchema]
    manufacturer_as_owner: Optional[bool]
    test_mode: Optional[bool]

    meta: Optional[MetaAPIData]

    class Config:
        extra = Extra.forbid

    def __str__(self):
        pstr = []

        uid = self.meta.id
        pstr.append(f"Creator: {uid}")

        nuuid = self.network.id
        pstr.append(f" - Network: {nuuid}")

        product = self.product
        pstr.append(f" - product: {product}")

        # Version 2.1 only
        if self.meta.version == WappstoVersion.v2_1:
            status = self.network.status.name
            pstr.append(f" - Status : {status}")

        return "\n".join(pstr)

    def _repr_pretty_(self, p, cycle):
        """
        Determents how the Creator Schema should be shown in IPython.

        # TODO: Add color check & colors.
        """
        print("\n" + str(self))  # NOTE: Error with WappstoVersion != v2_1


class SessionSchema(BaseModel):
    # TODO:
    provider: str
    type: str
    system: str

    remember_me: bool
    to_upgrade: bool
    upgrade: bool
    meta: MetaAPIData



import enum
import logging

from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

from .protocol import Connection
from ..utils import observer


class StatusID(str, enum.Enum):
    CONNECTING = "Connecting"
    CONNECTED = "Connected"
    DISCONNECTING = "Disconnecting"
    DISCONNETCED = "Disconnected"


class Debug(Connection):
    """For debuging purposes."""
    def __init__(self):
        self.log = logging.getLogger(__name__)
        self.log.addHandler(logging.NullHandler())

        self.observer_name = "CONNECTION"
        self.observer = observer
        self.observer.post(StatusID.DISCONNETCED, None)

        self.sended_data: Union[str, bytes, None] = None

        self.receive_data: List[Union[str, bytes]] = []

    def add_to_receive(self, data: Union[str, bytes]) -> None:
        self.receive_data.append(data)

    def send(
        self,
        data: Union[str, bytes]
    ) -> bool:
        """
        Send the str/Bytes to the server.

        If given string, it is encoded as 'uft-8' & send.

        Returns:
            True, if the data could be send else
            False.
        """
        self.sended_data = data
        return True

    def receive(
        self,
        parser: Callable[[bytes], Any]
    ) -> Any:
        """
        Socket receive method.

        Method that handles receiving data from a socket. Capable of handling
        data chunks.

        Args:
            Callable: A parser, that returns the parsed data.
                      On Parsing Error, it should raise a
                      ValueError TypeError or any subClasses of those.
                      (Like 'JSONDecodeError' & 'pydantic.ValidationError' is)
        Returns:
            The Parsers output.

        """
        return parser(self.receive_data.pop())

    def connect(self) -> bool:
        """
        Connect to the server.

        Attempts a connection to the server on the provided address and port.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        self.observer.post(StatusID.CONNECTING, None)

        self.observer.post(StatusID.CONNECTED, None)
        return True

    def reconnect(
        self,
        retry_limit: Optional[int] = None
    ) -> bool:
        """
        Attempt to reconnect.

        Close the current connection, and then try to reconnect to the server,
        until the given amount of attempts, are above the retry_limit.
        If the retry_limit are not set, it will continue end.

        Args:
            retry_limit: the amount of retries, before it stops.

        Returns:
            'True' if the connection was successful else
            'False'
        """
        self.log.info("Reconnection...")
        return True

    def disconnect(self) -> None:
        """Disconnect from the server."""
        pass

    def close(self) -> None:
        self.log.info("Closing connection...")
        self.observer.post(StatusID.DISCONNECTING, None)
        self.observer.post(StatusID.DISCONNETCED, None)


debug_connection = Debug()



import datetime


def timestamp() -> str:
    """
    Return The default timestamp used for Wappsto.

    The timestamp are always set to the UTC timezone.

    Returns:
        The UTC time string in ISO format.
    """
    return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')



import datetime
import logging
import random
import string

from enum import Enum

from typing import Optional
from typing import Union

import requests

from WappstoIoT.utils import Timestamp
from WappstoIoT.schema.base_schema import WappstoObject


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class TraceStatus(str, Enum):
    PENDING = "pending"  # On Receive
    OK = "ok"            # If Send Successful
    FAIL = "fail"        # If Send Failed.


ForcedTrace = False


def generateId() -> str:
    """Generate a Trace ID."""
    return "WappstoIoT_" + "".join(random.choices(
        string.ascii_letters + string.digits,
        k=10
    ))


def parentId(jsonrpc_elemt: Union[WappstoObject, dict]) -> Union[str, None]:
    """
    Check if a trace package should be send, if so it returns the parent ID.
    """
    log.debug(f"jsonrpc_elemt={jsonrpc_elemt}")
    if isinstance(jsonrpc_elemt, dict):
        return jsonrpc_elemt.get('params', {}).get('meta', {}).get('trace', None)
    return jsonrpc_elemt.meta.trace


def send(
    id: str,
    parent: str,
    name: str,
    status: TraceStatus,
    timestamp: Optional[datetime.datetime] = None
) -> Optional[dict]:
    """
    Send a Package Trace to Seluxit.

    Package tracing are used to debug, where the given package are lost,
    and/or the timing of the given package through the system.

    A trace package should be send when a JSONRPC object, with a trace value
    in the params, meta field, are received. The trace package should contain
    a status that are set to 'pending'.
    When the reply for the traced Wappsto json are ready to be send
    (to the socket). The Wappsto json's meta field should have the trace filed
    added, with the trace_id & name the pending trace package was send with,
    followed with the sending of another trace-package, where the status 'ok'.
    If for some reason it was not possible to generate a reply (but there
    should have been), the trace package should then be send with the
    'fail' status, on the realization this is the case.

    Args:
        trace_id: A generated ID, that should be added to
                  the Wappsto json meta trace filed.
        parent_id: The trace-value from the Wappsto json meta field.
        name: A descriptive name.
        status: Status for the traced package.
        timestamp: the timestamp in the ISO format.
                   If not sat, it will be sat in the time of sending.

    Returns:
        True, of the Trace package was send successful,
        False, if it was not.
    """
    params = {
        "id": id,
        "parent": parent,
        "name": name,
        "status": status,
        "timestamp": timestamp if timestamp else Timestamp.timestamp()
    }

    log.info(f"Trace id: {id}")

    r_data = requests.post(
        url='https://tracer.iot.seluxit.com/trace',
        params=params
    )
    log.debug(f"Trace reply: {r_data.text}")

    if r_data.status_code >= 300:
        log.error(f"Trace http error code: {r_data.status_code}")
        return None

    log.debug("Trace send!")

    return params



# import re
import ssl
import uuid
# import urllib

# from urllib.parse import ParseResult


class CertificateRead:
    def __init__(self, crt, ca=None):
        self.crt = ssl._ssl._test_decode_cert(crt)
        if ca:
            self.ca = ssl._ssl._test_decode_cert(ca)

    @property
    def network(self):
        return uuid.UUID(self.crt["subject"][0][0][1])

    @property
    def creator(self):
        return uuid.UUID(self.crt["subject"][1][0][1])

    @property
    def endpoint(self) -> str:
        return self.crt['issuer'][0][0][1]

        # return urllib.parse.urlparse(addr)

    @property
    def serial(self):
        return self.crt['serialNumber']

    @property
    def version(self):
        return self.crt['version']

    @property
    def notAfter(self):
        return self.crt['notAfter']



import logging
import threading

from typing import Any
from typing import Dict
from typing import List
from typing import Callable


obs_log = logging.getLogger(__name__)
obs_log.addHandler(logging.NullHandler())

default_subscriber = [
    lambda event_name, data: obs_log.debug(
        f"Default Observer: event_name={event_name}, data={data}"
    )
]

subscriber: Dict[str, List[Callable[[str, Any], None]]] = {}


def subscribe(event_name: str, callback: Callable[[str, Any], None]) -> None:
    """
    Subscribe to the event with given event name.

    Note: If lambda was used to subscribe with, it need to be the same
    instance that is used to unsubscribe with. So if it is needed to
    unsubscibe, save the instance.

    Args:
        event_name: The Unique name for the wanted event.
        callback: The function that need triggeret on the given event.
            The function will be called with the 'event_name', and 'data',
            that the event generate.
    """
    obs_log.debug(f"New Subscriber: {event_name}, {callback}")
    if event_name not in subscriber:
        subscriber[event_name] = default_subscriber.copy()
    subscriber[event_name].append(callback)


def post(event_name: str, data: Any) -> None:
    """
    Post the event with given event name.

    Args:
        event_name: An unique name for the given event.
        data: The given event subscriber might want.
    """
    def executer():
        for fn in subscriber.get(event_name, default_subscriber):
            fn(event_name, data)

    # NOTE: Needed for ensure that the receive thread do not get blocked.
    th = threading.Thread(target=executer)
    th.start()


def unsubscribe(event_name: str, callback: Callable[[str, Any], None]) -> bool:
    """
    Unsubscribe from given event name.

    Note: if lambda was used to subscribe with, it need to be the same
    instance that is used to unsubscribe with.

    Args:
        event_name: The Unique name for the wanted event.
        callback: The function that need triggeret on the given event.
            The function will be called with the 'event_name', and 'data',
            that the event generate.

    Returns:
        True, is the function was removed from the subcriber list.
        False, if it could not be, as in could not find it.
    """
    obs_log.debug(f"Unsubscribing: {event_name}, {callback}")

    try:
        # NOTE: Faster then testing if callback is in the list, when it is (60%+).
        subscriber.get(event_name, []).remove(callback)
    except ValueError:
        return False
    else:
        return True
